@@ -40,12 +40,13 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include <ldns/buffer.h>
 #include "types-internal.h"
 #include "util-internal.h"
 #include "dict.h"
 #include "rr-dict.h"
 #include "const-info.h"
+#include "gldns/gbuffer.h"
+#include "gldns/wire2str.h"
 
 /*---------------------------------------- getdns_dict_find */
 /**
@@ -511,7 +512,7 @@ priv_getdns_bindata_is_dname(struct getdns_bindata *bindata)
 
 /*---------------------------------------- getdns_pp_bindata */
 /**
- * private function to pretty print bindata to a ldns_buffer
+ * private function to pretty print bindata to a gldns_buffer
  * @param buf     buffer to write to
  * @param indent  number of spaces to append after newline
  * @param bindata the bindata to print
@@ -519,65 +520,91 @@ priv_getdns_bindata_is_dname(struct getdns_bindata *bindata)
  *                if an output error is encountered, a negative value
  */
 static int
-getdns_pp_bindata(ldns_buffer * buf, size_t indent,
-	struct getdns_bindata *bindata)
+getdns_pp_bindata(gldns_buffer *buf, size_t indent,
+	getdns_bindata *bindata, int rdata_raw, int json)
 {
-	size_t i, p = ldns_buffer_position(buf);
+	size_t i, p = gldns_buffer_position(buf);
 	uint8_t *dptr;
-	char *dname;
+	char spc[1024];
 
-	if (ldns_buffer_printf(buf, " <bindata ") < 0)
+	if (!json && gldns_buffer_printf(buf, " <bindata ") < 0)
 		return -1;
 
 	/* Walk through all printable characters */
 	i = 0;
-	if (bindata->size && bindata->data[bindata->size - 1] == 0)
-		while (i < bindata->size - 1 && isprint(bindata->data[i]))
+	if (!rdata_raw)
+		while (i < bindata->size && isprint(bindata->data[i]))
 			i++;
 
-	if (bindata->size > 1 && i >= bindata->size - 1) { /* all printable? */
-		if (ldns_buffer_printf(buf, "of \"%s\">", bindata->data) < 0)
+	if (bindata->size > 0 && i == bindata->size) {     /* all printable? */
+
+		if (json)
+			(void)snprintf(spc, sizeof(spc), "\"%%.%ds\"", (int)i);
+		else
+			(void)snprintf(spc, sizeof(spc), "of \"%%.%ds\"%s>",
+			    (int)(i > 32 ? 32 : i), (i > 32 ? "..." : ""));
+		if (gldns_buffer_printf(buf, spc, bindata->data) < 0)
+			return -1;
+
+	} else if (bindata->size > 1 &&         /* null terminated printable */
+	    i == bindata->size - 1 && bindata->data[i] == 0) {
+
+		if (gldns_buffer_printf(
+		    buf, (json ? "\"%s\"" : "of \"%s\">"), bindata->data) < 0)
 			return -1;
 
 	} else if (bindata->size == 1 && *bindata->data == 0) {
-		if (ldns_buffer_printf(buf, "for .>") < 0)
+		if (gldns_buffer_printf(buf, json ? "\".\"" : "for .>") < 0)
 			return -1;
 
 	} else if (priv_getdns_bindata_is_dname(bindata)) {
-		if (GETDNS_RETURN_GOOD ==
-			getdns_convert_dns_name_to_fqdn(bindata, &dname) &&
-			ldns_buffer_printf(buf, "for %s>", dname) < 0) {
-			free(dname);
+		(void)gldns_wire2str_dname_buf(
+		    bindata->data, bindata->size, spc, sizeof(spc));
+		if (gldns_buffer_printf(
+		    buf, (json ? "\"%s\"" : "of \"%s\">"), spc) < 0)
 			return -1;
+	} else if (json) {
+		if (gldns_buffer_printf(buf, "[") < 0)
+			return -1;
+		for (dptr = bindata->data;
+			dptr < bindata->data + bindata->size; dptr++) {
+			if (dptr > bindata->data) {
+				if (gldns_buffer_printf(buf, ",") < 0)
+					return -1;
+			}
+			if (gldns_buffer_printf(buf, "%d", (int)*dptr) < 0)
+				return -1;
 		}
-		free(dname);
-
+		if (gldns_buffer_printf(buf, "]") < 0)
+			return -1;
 	} else {
-		if (ldns_buffer_printf(buf, "of 0x") < 0)
+		if (gldns_buffer_printf(buf, "of 0x") < 0)
 			return -1;
 		for (dptr = bindata->data;
 			dptr < bindata->data + bindata->size; dptr++) {
 			if (dptr - bindata->data >= 16) {
-				if (ldns_buffer_printf(buf, "...") < 0)
+				if (gldns_buffer_printf(buf, "...") < 0)
 					return -1;
 				break;
 			}
-			if (ldns_buffer_printf(buf, "%.2x", *dptr) < 0)
+			if (gldns_buffer_printf(buf, "%.2x", *dptr) < 0)
 				return -1;
 		}
-		if (ldns_buffer_printf(buf, ">") < 0)
+		if (gldns_buffer_printf(buf, ">") < 0)
 			return -1;
 	}
-	return ldns_buffer_position(buf) - p;
+	return gldns_buffer_position(buf) - p;
 }				/* getdns_pp_bindata */
 
 static int
-getdns_pp_dict(ldns_buffer * buf, size_t indent,
-	const struct getdns_dict *dict);
+getdns_pp_dict(gldns_buffer * buf, size_t indent,
+	const getdns_dict *dict, int json);
+
+static const char *unknown_str_l[] = {" <unknown>", " null", "null"};
 
 /*---------------------------------------- getdns_pp_list */
 /**
- * private function to pretty print list to a ldns_buffer
+ * private function to pretty print list to a gldns_buffer
  * @param buf    buffer to write to
  * @param indent number of spaces to append after newline
  * @param list   the to list print
@@ -587,10 +614,10 @@ getdns_pp_dict(ldns_buffer * buf, size_t indent,
  *               if an output error is encountered, a negative value
  */
 static int
-getdns_pp_list(ldns_buffer *buf, size_t indent, getdns_list *list,
-    int for_namespaces)
+getdns_pp_list(gldns_buffer *buf, size_t indent, const getdns_list *list,
+    int for_namespaces, int json)
 {
-	size_t i, length, p = ldns_buffer_position(buf);
+	size_t i, length, p = gldns_buffer_position(buf);
 	getdns_data_type dtype;
 	struct getdns_dict *dict_item;
 	struct getdns_list *list_item;
@@ -601,7 +628,7 @@ getdns_pp_list(ldns_buffer *buf, size_t indent, getdns_list *list,
 	if (list == NULL)
 		return 0;
 
-	if (ldns_buffer_printf(buf, "[") < 0)
+	if (gldns_buffer_printf(buf, "[") < 0)
 		return -1;
 
 	if (getdns_list_get_length(list, &length) != GETDNS_RETURN_GOOD)
@@ -609,12 +636,14 @@ getdns_pp_list(ldns_buffer *buf, size_t indent, getdns_list *list,
 
 	indent += 2;
 	for (i = 0; i < length; i++) {
-		if (ldns_buffer_printf(buf, "%s\n%s", (i ? "," : ""),
-			getdns_indent(indent)) < 0)
+		if (i && gldns_buffer_printf(buf, ",") < 0)
+			return -1;
+		if (json < 2 && gldns_buffer_printf(buf, "\n%s",
+		    getdns_indent(indent)) < 0)
 			return -1;
 
-		if (getdns_list_get_data_type(list, i,
-			&dtype) != GETDNS_RETURN_GOOD)
+		if (getdns_list_get_data_type(list, i, &dtype) !=
+		    GETDNS_RETURN_GOOD)
 			return -1;
 
 		switch (dtype) {
@@ -622,29 +651,31 @@ getdns_pp_list(ldns_buffer *buf, size_t indent, getdns_list *list,
 
 			if (getdns_list_get_int(list, i, &int_item))
 				return -1;
-			if (for_namespaces &&
+			if (!json && for_namespaces &&
 			    (strval =
 			     priv_getdns_get_const_info(int_item)->name)) {
-				if (ldns_buffer_printf(buf, "%s", strval) < 0)
+				if (gldns_buffer_printf(buf, "%s", strval) < 0)
 					return -1;
 			} else if (0 > 
-			    ldns_buffer_printf(buf, "%d", (int)int_item))
+			    gldns_buffer_printf(buf, "%d", (int)int_item))
 				return -1;
 			break;
 
 		case t_bindata:
 			if (getdns_list_get_bindata(list, i, &bindata_item) !=
-				GETDNS_RETURN_GOOD)
+			    GETDNS_RETURN_GOOD)
 				return -1;
-			if (getdns_pp_bindata(buf, indent, bindata_item) < 0)
+			if (getdns_pp_bindata(
+			    buf, indent, bindata_item, 0, json) < 0)
 				return -1;
 			break;
 
 		case t_list:
 			if (getdns_list_get_list(list, i, &list_item) !=
-				GETDNS_RETURN_GOOD)
+			    GETDNS_RETURN_GOOD)
 				return -1;
-			if (getdns_pp_list(buf, indent, list_item, 0) < 0)
+			if (getdns_pp_list(
+			    buf, indent, list_item, 0, json) < 0)
 				return -1;
 			break;
 
@@ -652,71 +683,74 @@ getdns_pp_list(ldns_buffer *buf, size_t indent, getdns_list *list,
 			if (getdns_list_get_dict(list, i, &dict_item) !=
 				GETDNS_RETURN_GOOD)
 				return -1;
-			if (getdns_pp_dict(buf, indent, dict_item) < 0)
+			if (getdns_pp_dict(buf, indent, dict_item, json) < 0)
 				return -1;
 			break;
 
 		default:
-			if (ldns_buffer_printf(buf, " <unknown>") < 0)
+			if (gldns_buffer_printf(
+			    buf, "%s", unknown_str_l[json]) < 0)
 				return -1;
 		}
 	}
 	indent -= 2;
-	if (ldns_buffer_printf(buf, i ? "\n%s]" : "]",
-		getdns_indent(indent)) < 0)
+	if (json < 2 && i && gldns_buffer_printf(
+	    buf, "\n%s", getdns_indent(indent)) < 0)
+		return -1;
+	if (gldns_buffer_printf(buf, "]") < 0)
 		return -1;
 
-	return ldns_buffer_position(buf) - p;
+	return gldns_buffer_position(buf) - p;
 }				/* getdns_pp_list */
 
 static int
-priv_getdns_print_class(ldns_buffer *buf, uint32_t klass)
+priv_getdns_print_class(gldns_buffer *buf, uint32_t klass)
 {
 	switch (klass) {
 	case GETDNS_RRCLASS_IN:
-		(void) ldns_buffer_printf(buf, " GETDNS_RRCLASS_IN");
+		(void) gldns_buffer_printf(buf, " GETDNS_RRCLASS_IN");
 		return 1;
 	case GETDNS_RRCLASS_CH:
-		(void) ldns_buffer_printf(buf, " GETDNS_RRCLASS_CH");
+		(void) gldns_buffer_printf(buf, " GETDNS_RRCLASS_CH");
 		return 1;
 	case GETDNS_RRCLASS_HS:
-		(void) ldns_buffer_printf(buf, " GETDNS_RRCLASS_HS");
+		(void) gldns_buffer_printf(buf, " GETDNS_RRCLASS_HS");
 		return 1;
 	case GETDNS_RRCLASS_NONE:
-		(void) ldns_buffer_printf(buf, " GETDNS_RRCLASS_NONE");
+		(void) gldns_buffer_printf(buf, " GETDNS_RRCLASS_NONE");
 		return 1;
 	case GETDNS_RRCLASS_ANY:
-		(void) ldns_buffer_printf(buf, " GETDNS_RRCLASS_ANY");
+		(void) gldns_buffer_printf(buf, " GETDNS_RRCLASS_ANY");
 		return 1;
 	}
 	return 0;
 }
 
 static int
-priv_getdns_print_opcode(ldns_buffer *buf, uint32_t opcode)
+priv_getdns_print_opcode(gldns_buffer *buf, uint32_t opcode)
 {
 	switch (opcode) {
 	case GETDNS_OPCODE_QUERY:
-		(void) ldns_buffer_printf(buf, " GETDNS_OPCODE_QUERY");
+		(void) gldns_buffer_printf(buf, " GETDNS_OPCODE_QUERY");
 		return 1;
 	case GETDNS_OPCODE_IQUERY:
-		(void) ldns_buffer_printf(buf, " GETDNS_OPCODE_IQUERY");
+		(void) gldns_buffer_printf(buf, " GETDNS_OPCODE_IQUERY");
 		return 1;
 	case GETDNS_OPCODE_STATUS:
-		(void) ldns_buffer_printf(buf, " GETDNS_OPCODE_STATUS");
+		(void) gldns_buffer_printf(buf, " GETDNS_OPCODE_STATUS");
 		return 1;
 	case GETDNS_OPCODE_NOTIFY:
-		(void) ldns_buffer_printf(buf, " GETDNS_OPCODE_NOTIFY");
+		(void) gldns_buffer_printf(buf, " GETDNS_OPCODE_NOTIFY");
 		return 1;
 	case GETDNS_OPCODE_UPDATE:
-		(void) ldns_buffer_printf(buf, " GETDNS_OPCODE_UPDATE");
+		(void) gldns_buffer_printf(buf, " GETDNS_OPCODE_UPDATE");
 		return 1;
 	}
 	return 0;
 }
 
 static int
-priv_getdns_print_rcode(ldns_buffer *buf, uint32_t rcode)
+priv_getdns_print_rcode(gldns_buffer *buf, uint32_t rcode)
 {
 	static const char *rcodes[] = {
 		" GETDNS_RCODE_NOERROR" , " GETDNS_RCODE_FORMERR" ,
@@ -731,9 +765,9 @@ priv_getdns_print_rcode(ldns_buffer *buf, uint32_t rcode)
 		" GETDNS_RCODE_BADTRUNC"
 	};
 	if (rcode <= 10)
-		(void) ldns_buffer_printf(buf, rcodes[rcode]);
+		(void) gldns_buffer_printf(buf, "%s", rcodes[rcode]);
 	else if (rcode >= 16 && rcode <= 22)
-		(void) ldns_buffer_printf(buf, rcodes[rcode-6]);
+		(void) gldns_buffer_printf(buf, "%s", rcodes[rcode-6]);
 	else
 		return 0;
 	return 1;
@@ -741,7 +775,7 @@ priv_getdns_print_rcode(ldns_buffer *buf, uint32_t rcode)
 
 /*---------------------------------------- getdns_pp_dict */
 /**
- * private function to pretty print dict to a ldns_buffer
+ * private function to pretty print dict to a gldns_buffer
  * @param buf    buffer to write to
  * @param indent number of spaces to append after newline
  * @param dict   the dict to print
@@ -749,10 +783,10 @@ priv_getdns_print_rcode(ldns_buffer *buf, uint32_t rcode)
  *               if an output error is encountered, a negative value
  */
 static int
-getdns_pp_dict(ldns_buffer * buf, size_t indent,
-	const struct getdns_dict *dict)
+getdns_pp_dict(gldns_buffer * buf, size_t indent,
+	const getdns_dict *dict, int json)
 {
-	size_t i, length, p = ldns_buffer_position(buf);
+	size_t i, length, p = gldns_buffer_position(buf);
 	struct getdns_dict_item *item;
 	const char *strval;
 
@@ -761,30 +795,37 @@ getdns_pp_dict(ldns_buffer * buf, size_t indent,
 	if (dict == NULL)
 		return 0;
 
-	if (ldns_buffer_printf(buf, "{") < 0)
+	if (gldns_buffer_printf(buf, "{") < 0)
 		return -1;
 
 	i = 0;
 	indent += 2;
 	RBTREE_FOR(item, struct getdns_dict_item *,
 		(getdns_rbtree_t *)&(dict->root)) {
-		if (ldns_buffer_printf(buf, "%s\n%s\"%s\":", (i ? "," : "")
-			, getdns_indent(indent)
-			, item->node.key) < 0)
+
+		if (i && gldns_buffer_printf(buf, ",") < 0)
+			return -1;
+		if (json < 2 && gldns_buffer_printf(
+		    buf, "\n%s", getdns_indent(indent)) < 0)
+			return -1;
+		if (gldns_buffer_printf(
+		    buf, "\"%s\":",(const char *)item->node.key) < 0)
 			return -1;
 
 		switch (item->dtype) {
 		case t_int:
-			if ((strcmp(item->node.key, "type") == 0  ||
+			if (!json &&
+			    (strcmp(item->node.key, "type") == 0  ||
 			     strcmp(item->node.key, "type_covered") == 0 ||
 			     strcmp(item->node.key, "qtype") == 0) &&
 			    (strval = priv_getdns_rr_type_name(item->data.n))) {
-				if (ldns_buffer_printf(
+				if (gldns_buffer_printf(
 					buf, " GETDNS_RRTYPE_%s", strval) < 0)
 					return -1;
 				break;
 			}
-			if ((strcmp(item->node.key, "answer_type") == 0  ||
+			if (!json &&
+			    (strcmp(item->node.key, "answer_type") == 0  ||
 			     strcmp(item->node.key, "dnssec_status") == 0 ||
 			     strcmp(item->node.key, "status") == 0 ||
 			     strcmp(item->node.key, "append_name") == 0 ||
@@ -793,21 +834,23 @@ getdns_pp_dict(ldns_buffer * buf, size_t indent,
 			     strcmp(item->node.key, "resolution_type") == 0) &&
 			    (strval =
 			     priv_getdns_get_const_info(item->data.n)->name)) {
-				if (ldns_buffer_printf(buf, " %s", strval) < 0)
+				if (gldns_buffer_printf(buf, " %s", strval) < 0)
 					return -1;
 				break;
 			}
-			if ((strcmp(item->node.key, "class")  == 0  ||
-				 strcmp(item->node.key, "qclass") == 0) &&
-				priv_getdns_print_class(buf, item->data.n))
+			if (!json &&
+			    (strcmp(item->node.key, "class")  == 0  ||
+			     strcmp(item->node.key, "qclass") == 0) &&
+			    priv_getdns_print_class(buf, item->data.n))
 				break;
-			if (strcmp(item->node.key, "opcode") == 0 &&
-				priv_getdns_print_opcode(buf, item->data.n))
+			if (!json && strcmp(item->node.key, "opcode") == 0 &&
+			    priv_getdns_print_opcode(buf, item->data.n))
 				break;
-			if (strcmp(item->node.key, "rcode") == 0 &&
-				priv_getdns_print_rcode(buf, item->data.n))
+			if (!json && strcmp(item->node.key, "rcode") == 0 &&
+			    priv_getdns_print_rcode(buf, item->data.n))
 				break;
-			if (ldns_buffer_printf(buf, " %d", item->data.n) < 0)
+			if (gldns_buffer_printf(
+			    buf,(json < 2 ? " %d" : "%d"), item->data.n) < 0)
 				return -1;
 			break;
 
@@ -818,58 +861,67 @@ getdns_pp_dict(ldns_buffer * buf, size_t indent,
 			    (item->data.bindata->size == 4  ||
 			     item->data.bindata->size == 16 )) {
 
-				if (ldns_buffer_printf(buf, " <bindata for %s>",
+				if (gldns_buffer_printf(buf, 
+				    (json ? "\"%s\"" : " <bindata for %s>"),
 				    inet_ntop(( item->data.bindata->size == 4
 				              ? AF_INET : AF_INET6)
-					     , item->data.bindata->data
-					     , abuf
-					     , 40
-					     )) < 0)
+				              , item->data.bindata->data
+				              , abuf
+				              , 40
+				              )) < 0)
 					return -1;
 	
-			} else if (getdns_pp_bindata(buf, indent,
-			    item->data.bindata) < 0)
+			} else if (getdns_pp_bindata(
+			    buf, indent, item->data.bindata,
+			    (strcmp(item->node.key, "rdata_raw") == 0),
+			    json) < 0)
 				return -1;
 			break;
 
 		case t_list:	/* Don't put empty lists on a new line */
 
-			if (getdns_list_get_length(item->data.list,
-				&length) != GETDNS_RETURN_GOOD)
+			if (getdns_list_get_length(item->data.list, &length) !=
+			    GETDNS_RETURN_GOOD)
 				return -1;
 			if (length == 0) {
-				if (ldns_buffer_printf(buf, " []") < 0)
+				if (gldns_buffer_printf(
+				    buf, (json < 2 ? " []" : "[]")) < 0)
 					return -1;
 				break;
 			}
-			if (ldns_buffer_printf(buf, "\n%s",
-				getdns_indent(indent)) < 0)
+			if (json < 2 && gldns_buffer_printf(buf, "\n%s",
+			    getdns_indent(indent)) < 0)
 				return -1;
 			if (getdns_pp_list(buf, indent, item->data.list, 
-			    (strcmp(item->node.key, "namespaces") == 0)) < 0)
+			    (strcmp(item->node.key, "namespaces") == 0),
+			    json) < 0)
 				return -1;
 			break;
 
 		case t_dict:
-			if (ldns_buffer_printf(buf, "\n%s",
-				getdns_indent(indent)) < 0)
+			if (json < 2 && gldns_buffer_printf(buf, "\n%s",
+			    getdns_indent(indent)) < 0)
 				return -1;
-			if (getdns_pp_dict(buf, indent, item->data.dict) < 0)
+			if (getdns_pp_dict(
+			    buf, indent, item->data.dict, json) < 0)
 				return -1;
 			break;
 
 		default:
-			if (ldns_buffer_printf(buf, " <unknown>") < 0)
+			if (gldns_buffer_printf(
+			    buf, "%s", unknown_str_l[json]) < 0)
 				return -1;
 		}
 		i++;
 	}
 	indent -= 2;
-	if (ldns_buffer_printf(buf, i ? "\n%s}" : "}",
-		getdns_indent(indent)) < 0)
+	if (json < 2 && i  && gldns_buffer_printf(
+	    buf, "\n%s", getdns_indent(indent)) < 0)
+		return -1;
+	if (gldns_buffer_printf(buf, "}") < 0)
 		return -1;
 
-	return ldns_buffer_position(buf) - p;
+	return gldns_buffer_position(buf) - p;
 }				/* getdns_pp_dict */
 
 /*---------------------------------------- getdns_pretty_print_dict */
@@ -883,25 +935,141 @@ getdns_pp_dict(ldns_buffer * buf, size_t indent,
 char *
 getdns_pretty_print_dict(const struct getdns_dict *dict)
 {
-	ldns_buffer *buf;
+	gldns_buffer *buf;
 	char *ret;
 
 	if (!dict)
 		return NULL;
 
-	buf = ldns_buffer_new(100);
+	buf = gldns_buffer_new(8192);
 	if (!buf)
 		return NULL;
 
-	if (getdns_pp_dict(buf, 0, dict) < 0) {
-		ldns_buffer_free(buf);
+	if (getdns_pp_dict(buf, 0, dict, 0) < 0) {
+		gldns_buffer_free(buf);
 		return NULL;
 	}
-	ret = (char *) ldns_buffer_export(buf);
-	ldns_buffer_free(buf);
+	ret = (char *) gldns_buffer_export(buf);
+	gldns_buffer_free(buf);
 	return ret;
 }				/* getdns_pretty_print_dict */
 
+int
+getdns_pretty_snprint_dict(char *str, size_t size, const getdns_dict *dict)
+{
+	gldns_buffer buf;
+
+	if (!dict) return -1;
+
+	gldns_buffer_init_frm_data(&buf, str, size);
+	return getdns_pp_dict(&buf, 0, dict, 0) < 0
+	     ? -1 : gldns_buffer_position(&buf);
+}
+
+char *
+getdns_pretty_print_list(const getdns_list *list)
+{
+	gldns_buffer *buf;
+	char *ret;
+
+	if (!list)
+		return NULL;
+
+	buf = gldns_buffer_new(4096);
+	if (!buf)
+		return NULL;
+
+	if (getdns_pp_list(buf, 0, list, 0, 0) < 0) {
+		gldns_buffer_free(buf);
+		return NULL;
+	}
+	ret = (char *) gldns_buffer_export(buf);
+	gldns_buffer_free(buf);
+	return ret;
+}
+
+int
+getdns_pretty_snprint_list(char *str, size_t size, const getdns_list *list)
+{
+	gldns_buffer buf;
+
+	if (!list) return -1;
+
+	gldns_buffer_init_frm_data(&buf, str, size);
+	return getdns_pp_list(&buf, 0, list, 0, 0) < 0
+	     ? -1 : gldns_buffer_position(&buf);
+}
+
+char *
+getdns_print_json_dict(const getdns_dict *dict, int pretty)
+{
+	gldns_buffer *buf;
+	char *ret;
+
+	if (!dict)
+		return NULL;
+
+	buf = gldns_buffer_new(8192);
+	if (!buf)
+		return NULL;
+
+	if (getdns_pp_dict(buf, 0, dict, pretty ? 1 : 2) < 0) {
+		gldns_buffer_free(buf);
+		return NULL;
+	}
+	ret = (char *) gldns_buffer_export(buf);
+	gldns_buffer_free(buf);
+	return ret;
+}				/* getdns_print_json_dict */
+
+int
+getdns_snprint_json_dict(
+    char *str, size_t size, const getdns_dict *dict, int pretty)
+{
+	gldns_buffer buf;
+
+	if (!dict) return -1;
+
+	gldns_buffer_init_frm_data(&buf, str, size);
+	return getdns_pp_dict(&buf, 0, dict, pretty ? 1 : 2) < 0
+	     ? -1 : gldns_buffer_position(&buf);
+}
+
+char *
+getdns_print_json_list(const getdns_list *list, int pretty)
+{
+	gldns_buffer *buf;
+	char *ret;
+
+	if (!list)
+		return NULL;
+
+	buf = gldns_buffer_new(4096);
+	if (!buf)
+		return NULL;
+
+	if (getdns_pp_list(buf, 0, list, 0, pretty ? 1 : 2) < 0) {
+		gldns_buffer_free(buf);
+		return NULL;
+	}
+	ret = (char *) gldns_buffer_export(buf);
+	gldns_buffer_free(buf);
+	return ret;
+}
+
+int
+getdns_snprint_json_list(
+    char *str, size_t size, const getdns_list *list, int pretty)
+{
+	gldns_buffer buf;
+
+	if (!list) return -1;
+
+	gldns_buffer_init_frm_data(&buf, str, size);
+	return getdns_pp_list(&buf, 0, list, 0, pretty ? 1 : 2) < 0
+	     ? -1 : gldns_buffer_position(&buf);
+}
+
 getdns_return_t
 getdns_dict_remove_name(struct getdns_dict *this_dict, const char *name)
 {
