@@ -47,6 +47,9 @@
 #include "types-internal.h"
 #include "dnssec.h"
 #include "rr-dict.h"
+#include "gldns/str2wire.h"
+#include "gldns/wire2str.h"
+#include "general.h"
 
 void priv_getdns_call_user_callback(getdns_dns_req *, struct getdns_dict *);
 
@@ -60,11 +63,9 @@ struct validation_chain {
 
 struct chain_response {
 	int err;
-	ldns_rr_list *result;
-	int sec;
-	char *bogus;
+	getdns_list *result;
 	struct validation_chain *chain;
-	int unbound_id;
+	getdns_dns_req *dns_req;
 };
 
 struct chain_link {
@@ -73,17 +74,80 @@ struct chain_link {
 	struct chain_response DS;
 };
 
-static void launch_chain_link_lookup(struct validation_chain *chain, char *name);
+static void launch_chain_link_lookup(struct validation_chain *chain,
+    priv_getdns_rdf_iter *rdf_dname);
 static void destroy_chain(struct validation_chain *chain);
 
+#ifdef STUB_NATIVE_DNSSEC
+static void
+native_stub_validate_dnssec(getdns_dns_req *dns_req, getdns_list *support)
+{
+	getdns_network_req *netreq, **netreq_p;
+	getdns_list *trust_anchors;
+	getdns_dict *reply = NULL;
+	getdns_list *to_validate;
+	getdns_list *list;
+	getdns_dict *rr_dict;
+	size_t i;
+
+	if (!(trust_anchors = getdns_root_trust_anchor(NULL)))
+		return;
+
+	for (netreq_p = dns_req->netreqs; (netreq = *netreq_p) ; netreq_p++) {
+		if (!(reply = priv_getdns_create_reply_dict(dns_req->context,
+		    netreq, NULL, NULL)))
+			continue;
+		if (!(to_validate =
+		    getdns_list_create_with_context(dns_req->context)))
+			break;
+		if (getdns_dict_get_list(reply, "answer", &list)) {
+			getdns_list_destroy(to_validate);
+			break;
+		}
+		for (i = 0; !getdns_list_get_dict(list, i, &rr_dict); i++)
+			(void) getdns_list_append_dict(to_validate, rr_dict);
+
+		if (getdns_dict_get_list(reply, "authority", &list)) {
+			getdns_list_destroy(to_validate);
+			break;
+		}
+		for (i = 0; !getdns_list_get_dict(list, i, &rr_dict); i++)
+			(void) getdns_list_append_dict(to_validate, rr_dict);
+
+		switch ((int)getdns_validate_dnssec(
+		    to_validate, support, trust_anchors)) {
+		case GETDNS_DNSSEC_SECURE:
+			netreq->secure = 1;
+			netreq->bogus  = 0;
+			break;
+		case GETDNS_DNSSEC_BOGUS:
+			netreq->secure = 0;
+			netreq->bogus  = 1;
+			break;
+		default:
+			/* GETDNS_DNSSEC_INSECURE */
+			netreq->secure = 0;
+			netreq->bogus  = 0;
+			break;
+		}
+		getdns_list_destroy(to_validate);
+		getdns_dict_destroy(reply);
+		reply = NULL;
+	}
+	getdns_list_destroy(trust_anchors);
+	getdns_dict_destroy(reply);
+}
+#endif
+
 static void callback_on_complete_chain(struct validation_chain *chain)
 {
-	struct getdns_context *context = chain->dns_req->context;
-	struct getdns_dict *response;
+	getdns_context *context = chain->dns_req->context;
+	getdns_dict *response;
 	struct chain_link *link;
 	size_t ongoing = chain->lock;
-	ldns_rr_list *keys;
-	struct getdns_list *getdns_keys;
+	getdns_list *keys;
+	size_t i;
+	getdns_dict *rr_dict;
 
 	RBTREE_FOR(link, struct chain_link *,
 	    (getdns_rbtree_t *)&(chain->root)) {
@@ -94,83 +158,112 @@ static void callback_on_complete_chain(struct validation_chain *chain)
 		    ((const char *)link->node.key)[1] != '\0' ))
 		       	ongoing++;
 	}
-	if (ongoing == 0) {
-		getdns_dns_req *dns_req = chain->dns_req;
-		response = create_getdns_response(chain->dns_req);
+	if (ongoing > 0)
+		return;
 
-		keys = ldns_rr_list_new();
-		RBTREE_FOR(link, struct chain_link *,
-		    (getdns_rbtree_t *)&(chain->root)) {
-			(void) ldns_rr_list_cat(keys, link->DNSKEY.result);
-			(void) ldns_rr_list_cat(keys, link->DS.result);
-		}
-		getdns_keys = create_list_from_rr_list(context, keys);
-		(void) getdns_dict_set_list(response, "validation_chain",
-		    getdns_keys);
-		getdns_list_destroy(getdns_keys);
-		ldns_rr_list_free(keys);
-		priv_getdns_call_user_callback(dns_req, response);
+	if (!(keys = getdns_list_create_with_context(context))) {
+		priv_getdns_call_user_callback(chain->dns_req,
+		    create_getdns_response(chain->dns_req));
 		destroy_chain(chain);
+		return;
 	}
+	RBTREE_FOR(link, struct chain_link *,
+	    (getdns_rbtree_t *)&(chain->root)) {
+		for (i = 0; !getdns_list_get_dict( link->DNSKEY.result
+						 , i, &rr_dict); i++)
+			(void) getdns_list_append_dict(keys, rr_dict);
+		for (i = 0; !getdns_list_get_dict( link->DS.result
+						 , i, &rr_dict); i++)
+			(void) getdns_list_append_dict(keys, rr_dict);
+	}
+#ifdef STUB_NATIVE_DNSSEC
+	native_stub_validate_dnssec(chain->dns_req, keys);
+#endif
+	if ((response = create_getdns_response(chain->dns_req)) &&
+	    chain->dns_req->dnssec_return_validation_chain) {
+	    (void)getdns_dict_set_list(response, "validation_chain", keys);
+	}
+	getdns_list_destroy(keys);
+	priv_getdns_call_user_callback(chain->dns_req, response);
+	destroy_chain(chain);
 }
 
 
 static void
-ub_chain_response_callback(void *arg, int err, struct ub_result* ub_res)
+chain_response_callback(struct getdns_dns_req *dns_req)
 {
-	struct chain_response *response = (struct chain_response *) arg;
-    ldns_status r;
-    ldns_pkt *p;
-    ldns_rr_list *answer;
-    ldns_rr_list *keys;
-    size_t i;
+	struct chain_response *response =
+	    (struct chain_response *) dns_req->user_pointer;
+	getdns_context *context = dns_req->context;
+	getdns_network_req **netreq_p, *netreq;
+	priv_getdns_rr_iter rr_iter_storage, *rr_iter;
+	priv_getdns_rdf_iter rdf_storage, *rdf;
+	gldns_pkt_section section;
+	uint16_t rr_type, type_covered;
+	getdns_dict *rr_dict;
+	getdns_list *keys;
+	size_t nkeys;
+	getdns_return_t r;
 
-    response->err    = err;
-    response->sec    = ub_res ? ub_res->secure : 0;
-    response->bogus  = ub_res ? ub_res->why_bogus : NULL;
+	response->dns_req = dns_req;
+	if (!(keys = getdns_list_create_with_context(context)))
+	    goto done;
 
-    if (ub_res == NULL)
-        goto done;
+	for (netreq_p = dns_req->netreqs; (netreq = *netreq_p); netreq_p++) {
+		for ( rr_iter = priv_getdns_rr_iter_init(&rr_iter_storage
+							, netreq->response
+							, netreq->response_len)
+		    ; rr_iter
+		    ; rr_iter = priv_getdns_rr_iter_next(rr_iter)
+		    ) {
+			section = priv_getdns_rr_iter_section(rr_iter);
+			if (section != GLDNS_SECTION_ANSWER)
+				continue;
 
-    r = ldns_wire2pkt(&p, ub_res->answer_packet, ub_res->answer_len);
-	if (r != LDNS_STATUS_OK) {
-		if (err == 0)
-			response->err = r;
-		goto done;
-	}
+			rr_type = gldns_read_uint16(rr_iter->rr_type);
 
-	keys = ldns_rr_list_new();
-	answer = ldns_pkt_answer(p);
-	for (i = 0; i < ldns_rr_list_rr_count(answer); i++) {
-		ldns_rr *rr = ldns_rr_list_rr(answer, i);
+			if (rr_type == GETDNS_RRTYPE_DS ||
+			    rr_type == GETDNS_RRTYPE_DNSKEY) {
+				if (!(rr_dict = priv_getdns_rr_iter2rr_dict(
+				    context, rr_iter)))
+					continue;
+				r = getdns_list_append_dict(keys, rr_dict);
+				getdns_dict_destroy(rr_dict);
+				if (r) break;
+			}
+			if (rr_type != GETDNS_RRTYPE_RRSIG)
+				continue;
 
-		if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_DNSKEY ||
-		    ldns_rr_get_type(rr) == LDNS_RR_TYPE_DS) {
+			if (!(rdf = priv_getdns_rdf_iter_init(
+			    &rdf_storage, rr_iter)))
+				continue;
 
-			(void) ldns_rr_list_push_rr(keys, ldns_rr_clone(rr));
-			continue;
+			type_covered = gldns_read_uint16(rdf->pos);
+			if (type_covered == GETDNS_RRTYPE_DS) {
+				if ((rdf = priv_getdns_rdf_iter_init_at(
+				    &rdf_storage, rr_iter, 7)))
+					launch_chain_link_lookup(
+					    response->chain, rdf);
+
+			} else if (type_covered != GETDNS_RRTYPE_DNSKEY)
+				continue;
+
+			if (!(rr_dict = priv_getdns_rr_iter2rr_dict(
+			    context, rr_iter)))
+				continue;
+			r = getdns_list_append_dict(keys, rr_dict);
+			getdns_dict_destroy(rr_dict);
+			if (r) break;
 		}
-		if (ldns_rr_get_type(rr) != LDNS_RR_TYPE_RRSIG)
-			continue;
-
-		if (ldns_read_uint16(ldns_rdf_data(ldns_rr_rdf(rr, 0))) ==
-		    LDNS_RR_TYPE_DS)
-			launch_chain_link_lookup(response->chain,
-			    ldns_rdf2str(ldns_rr_rdf(rr, 7)));
-
-		else if (ldns_read_uint16(ldns_rdf_data(ldns_rr_rdf(rr, 0))) !=
-		    LDNS_RR_TYPE_DNSKEY)
-			continue;
-
-		(void) ldns_rr_list_push_rr(keys, ldns_rr_clone(rr));
 	}
-	if (ldns_rr_list_rr_count(keys))
+	if (getdns_list_get_length(keys, &nkeys))
+		getdns_list_destroy(keys);
+
+	else if (!nkeys)
+		getdns_list_destroy(keys);
+	else
 		response->result = keys;
-	else
-		ldns_rr_list_free(keys);
 
-	ldns_pkt_free(p);
-    ub_resolve_free(ub_res);
 
 done:	if (response->err == 0 && response->result == NULL)
 		response->err = -1;
@@ -181,36 +274,58 @@ done:	if (response->err == 0 && response->result == NULL)
 static void chain_response_init(
     struct validation_chain *chain, struct chain_response *response)
 {
-	response->err        = 0;
-	response->result     = NULL;
-	response->sec        = 0;
-	response->bogus      = NULL;
-	response->chain      = chain;
-	response->unbound_id = -1;
+	response->err     = 0;
+	response->result  = NULL;
+	response->chain   = chain;
+	response->dns_req = NULL;
 }
 
 static int
 resolve(char* name, int rrtype, struct chain_response *response)
 {
-	return ub_resolve_async(
-	    response->chain->dns_req->context->unbound_ctx,
-	    name, rrtype, LDNS_RR_CLASS_IN, response,
-	    ub_chain_response_callback, &response->unbound_id);
+	getdns_return_t r;
+	getdns_dict *extensions;
+
+	if (!(extensions = getdns_dict_create_with_context(
+	    response->chain->dns_req->context)))
+		return GETDNS_RETURN_MEMORY_ERROR;
+
+	if (!(r = getdns_dict_set_int(extensions,
+	    "dnssec_ok_checking_disabled", GETDNS_EXTENSION_TRUE)))
+
+		r = priv_getdns_general_loop(response->chain->dns_req->context,
+		    response->chain->dns_req->loop, name, rrtype, extensions,
+		    response, NULL, NULL, chain_response_callback);
+
+	getdns_dict_destroy(extensions);
+	return r;
 }
 
 static void
-launch_chain_link_lookup(struct validation_chain *chain, char *name)
+launch_chain_link_lookup(
+    struct validation_chain *chain, priv_getdns_rdf_iter *rdf_dname)
 {
 	int r;
-	struct chain_link *link = (struct chain_link *)
-	    getdns_rbtree_search((getdns_rbtree_t *)&(chain->root), name);
-
-	if (link) {
-		free(name);
+	struct chain_link *link;
+	uint8_t dname_spc[256], *dname;
+	char name[1024];
+	size_t dname_spc_sz = sizeof(dname_spc);
+	
+	if (!(dname = priv_getdns_rdf_if_or_as_decompressed(
+	    rdf_dname, dname_spc, &dname_spc_sz)))
 		return;
-	}
+
+	if (!gldns_wire2str_dname_buf(dname, (dname == dname_spc ?
+	    sizeof(dname_spc) : rdf_dname->nxt - rdf_dname->pos),
+	    name, sizeof(name)))
+		return;
+
+	if ((link = (struct chain_link *)
+	    getdns_rbtree_search((getdns_rbtree_t *)&(chain->root), name)))
+		return;
+
 	link = GETDNS_MALLOC(chain->mf, struct chain_link);
-	link->node.key = name;
+	link->node.key = getdns_strdup(&chain->mf, name);
 
 	chain_response_init(chain, &link->DNSKEY);
 	chain_response_init(chain, &link->DS);
@@ -218,12 +333,12 @@ launch_chain_link_lookup(struct validation_chain *chain, char *name)
 	getdns_rbtree_insert(&(chain->root), (getdns_rbnode_t *)link);
 
 	chain->lock++;
-	r = resolve(name, LDNS_RR_TYPE_DNSKEY, &link->DNSKEY);
+	r = resolve(name, GETDNS_RRTYPE_DNSKEY, &link->DNSKEY);
 	if (r != 0)
 		link->DNSKEY.err = r;
 
 	if (name[0] != '.' || name[1] != '\0') {
-		r = resolve(name, LDNS_RR_TYPE_DS, &link->DS);
+		r = resolve(name, GETDNS_RRTYPE_DS, &link->DS);
 		if (r != 0)
 			link->DS.err = r;
 	}
@@ -257,8 +372,17 @@ static void destroy_chain_link(getdns_rbnode_t * node, void *arg)
 	struct validation_chain *chain   = (struct validation_chain*) arg;
 
 	free((void *)link->node.key);
-	ldns_rr_list_deep_free(link->DNSKEY.result);
-	ldns_rr_list_deep_free(link->DS.result);
+
+	getdns_list_destroy(link->DNSKEY.result);
+	if (link->DNSKEY.dns_req) {
+		getdns_context_clear_outbound_request(link->DNSKEY.dns_req);
+		dns_req_free(link->DNSKEY.dns_req);
+	}
+	getdns_list_destroy(link->DS.result);
+	if (link->DS.dns_req) {
+		getdns_context_clear_outbound_request(link->DS.dns_req);
+		dns_req_free(link->DS.dns_req);
+	}
 	GETDNS_FREE(chain->mf, link);
 }
 
@@ -272,31 +396,41 @@ static void destroy_chain(struct validation_chain *chain)
 static void
 getdns_get_validation_chain(getdns_dns_req *dns_req, uint64_t *timeout)
 {
-	getdns_network_req *netreq = dns_req->first_req;
+	getdns_network_req **netreq_p, *netreq;
 	struct validation_chain *chain = create_chain(dns_req, timeout);
+	priv_getdns_rr_iter rr_iter_storage, *rr_iter;
+	priv_getdns_rdf_iter rdf_storage, *rdf;
+	gldns_pkt_section section;
+	uint16_t rr_type;
 
 	if (! chain) {
 		priv_getdns_call_user_callback(
 		    dns_req, create_getdns_response(dns_req));
 		return;
 	}
-	while (netreq) {
-		size_t i;
-		ldns_rr_list *answer = ldns_pkt_answer(netreq->result);
-		ldns_rr_list *authority = ldns_pkt_authority(netreq->result);
-		for (i = 0; i < ldns_rr_list_rr_count(answer); i++) {
-			ldns_rr *rr = ldns_rr_list_rr(answer, i);
-			if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_RRSIG)
-				launch_chain_link_lookup(chain,
-				    ldns_rdf2str(ldns_rr_rdf(rr, 7)));
+	for (netreq_p = dns_req->netreqs; (netreq = *netreq_p); netreq_p++) {
+
+		for ( rr_iter = priv_getdns_rr_iter_init(&rr_iter_storage
+		                                        , netreq->response
+							, netreq->response_len)
+		    ; rr_iter
+		    ; rr_iter = priv_getdns_rr_iter_next(rr_iter)
+		    ) {
+			section = priv_getdns_rr_iter_section(rr_iter);
+			if (section != GLDNS_SECTION_ANSWER &&
+			    section != GLDNS_SECTION_AUTHORITY)
+				continue;
+
+			rr_type = gldns_read_uint16(rr_iter->rr_type);
+			if (rr_type != GETDNS_RRTYPE_RRSIG)
+				continue;
+
+			if (!(rdf =  priv_getdns_rdf_iter_init_at(
+			    &rdf_storage , rr_iter, 7)))
+				continue;
+
+			launch_chain_link_lookup(chain, rdf);
 		}
-		for (i = 0; i < ldns_rr_list_rr_count(authority); i++) {
-			ldns_rr *rr = ldns_rr_list_rr(authority, i);
-			if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_RRSIG)
-				launch_chain_link_lookup(chain,
-				    ldns_rdf2str(ldns_rr_rdf(rr, 7)));
-		}
-		netreq = netreq->next;
 	}
 	callback_on_complete_chain(chain);
 }
@@ -341,6 +475,84 @@ priv_getdns_rr_list_from_list(struct getdns_list *list, ldns_rr_list **rr_list)
 	return r;
 }
 
+static int
+priv_getdns_rr_dict_with_compressed_names(getdns_dict *rr_dict)
+{
+	uint32_t rr_type;
+	getdns_dict *rdata;
+
+	if (getdns_dict_get_int(rr_dict, "type", &rr_type))
+		return 0;
+	if (rr_type == GETDNS_RRTYPE_RRSIG) {
+		if (getdns_dict_get_dict(rr_dict, "rdata", &rdata))
+			return 0;
+		if (getdns_dict_get_int(rdata, "type_covered", &rr_type))
+			return 0;
+	}
+	switch (rr_type) {
+	case GETDNS_RRTYPE_NS:
+	case GETDNS_RRTYPE_MD:
+	case GETDNS_RRTYPE_CNAME:
+	case GETDNS_RRTYPE_SOA:
+	case GETDNS_RRTYPE_MG:
+	case GETDNS_RRTYPE_MR:
+	case GETDNS_RRTYPE_PTR:
+	case GETDNS_RRTYPE_MINFO:
+	case GETDNS_RRTYPE_MX:
+		return 1;
+	default:
+		return 0;
+	}
+}
+
+static int
+ldns_dname_compare_v(const void *a, const void *b) {
+	return ldns_dname_compare((ldns_rdf *)a, (ldns_rdf *)b);
+}
+
+ldns_status
+priv_getdns_ldns_dnssec_zone_add_rr(ldns_dnssec_zone *zone, ldns_rr *rr)
+{
+	ldns_dnssec_name *new_name;
+	ldns_rbnode_t *new_node;
+
+	if (ldns_rr_get_type(rr) != LDNS_RR_TYPE_NSEC3)
+		return ldns_dnssec_zone_add_rr(zone, rr);
+	
+	if (!(new_name = ldns_dnssec_name_new()))
+		return LDNS_STATUS_MEM_ERR;
+
+	new_name->name = ldns_rdf_clone(ldns_rr_owner(rr));
+	new_name->hashed_name = ldns_dname_label(ldns_rr_owner(rr), 0);
+	new_name->name_alloced = true;
+
+	if (!(new_node = LDNS_MALLOC(ldns_rbnode_t))) {
+		ldns_dnssec_name_free(new_name);
+		return LDNS_STATUS_MEM_ERR;
+	}
+	new_node->key = new_name->name;
+	new_node->data = new_name;
+	if (!zone->names)
+		zone->names = ldns_rbtree_create(ldns_dname_compare_v);
+	(void)ldns_rbtree_insert(zone->names, new_node);
+
+#ifdef LDNS_DNSSEC_ZONE_HASHED_NAMES
+	if (!(new_node = LDNS_MALLOC(ldns_rbnode_t))) {
+		ldns_dnssec_name_free(new_name);
+		return LDNS_STATUS_MEM_ERR;
+	}
+	new_node->key = new_name->hashed_name;
+	new_node->data = new_name;
+	if (!zone->hashed_names) {
+		zone->_nsec3params = rr;
+		zone->hashed_names = ldns_rbtree_create(ldns_dname_compare_v);
+	}
+	(void)ldns_rbtree_insert(zone->hashed_names, new_node);
+#endif
+
+	return ldns_dnssec_zone_add_rr(zone, rr);
+}
+
 static getdns_return_t
 priv_getdns_dnssec_zone_from_list(struct getdns_list *list,
     ldns_dnssec_zone **zone)
@@ -361,10 +573,13 @@ priv_getdns_dnssec_zone_from_list(struct getdns_list *list,
 		if ((r = getdns_list_get_dict(list, i, &rr_dict)))
 			break;
 
+		if (priv_getdns_rr_dict_with_compressed_names(rr_dict))
+			continue;
+
 		if ((r = priv_getdns_create_rr_from_dict(rr_dict, &rr)))
 			break;
 
-		if ((s = ldns_dnssec_zone_add_rr(*zone, rr))) {
+		if ((s = priv_getdns_ldns_dnssec_zone_add_rr(*zone, rr))) {
 			ldns_rr_free(rr);
 			r = GETDNS_RETURN_GENERIC_ERROR;
 			break;
@@ -379,19 +594,41 @@ typedef struct zone_iter {
 	ldns_dnssec_zone   *zone;
 	ldns_rbnode_t      *cur_node;
 	ldns_dnssec_rrsets *cur_rrset;
+
+	ldns_dnssec_rrsets nsec_rrset;
+	ldns_dnssec_rrs    nsec_rrs;
 } zone_iter;
 
 static void
 rrset_iter_init_zone(zone_iter *i, ldns_dnssec_zone *zone)
 {	
+	ldns_dnssec_name *name;
 	assert(i);
 
 	i->zone = zone;
-	i->cur_node = zone->names ? ldns_rbtree_first(zone->names)
-	                          : LDNS_RBTREE_NULL;
-	i->cur_rrset = i->cur_node != LDNS_RBTREE_NULL
-	    ? ((ldns_dnssec_name *)i->cur_node->data)->rrsets
-	    : NULL;
+	if ((i->cur_node = zone->names
+	                 ? ldns_rbtree_first(zone->names)
+	                 : LDNS_RBTREE_NULL) == LDNS_RBTREE_NULL) {
+
+		i->cur_rrset = NULL;
+		return;
+	}
+
+	i->cur_rrset = ((ldns_dnssec_name *)i->cur_node->data)->rrsets;
+	if (!i->cur_rrset) {
+		name = ((ldns_dnssec_name *)i->cur_node->data);
+		if (name->nsec && name->nsec_signatures) {
+			i->cur_rrset = &i->nsec_rrset;
+			i->nsec_rrset.rrs = &i->nsec_rrs;
+			i->nsec_rrs.rr = name->nsec;
+			i->nsec_rrs.next = NULL;
+			i->nsec_rrset.type = ldns_rr_get_type(name->nsec);
+			i->nsec_rrset.signatures =
+				name->nsec_signatures;
+			i->nsec_rrset.next = NULL;
+			return;
+		}
+	}
 }
 
 static ldns_dnssec_rrsets *
@@ -405,12 +642,30 @@ rrset_iter_value(zone_iter *i)
 static void
 rrset_iter_next(zone_iter *i)
 {
+	int was_nsec_rrset;
+	ldns_dnssec_name *name;
 	assert(i);
 
 	if (! i->cur_rrset)
 		return;
 
+	was_nsec_rrset = (i->cur_rrset == &i->nsec_rrset);
 	if (!  (i->cur_rrset = i->cur_rrset->next)) {
+
+		if (!was_nsec_rrset) {
+			name = ((ldns_dnssec_name *)i->cur_node->data);
+			if (name->nsec && name->nsec_signatures) {
+				i->cur_rrset = &i->nsec_rrset;
+				i->nsec_rrset.rrs = &i->nsec_rrs;
+				i->nsec_rrs.rr = name->nsec;
+				i->nsec_rrs.next = NULL;
+				i->nsec_rrset.type = ldns_rr_get_type(name->nsec);
+				i->nsec_rrset.signatures =
+					name->nsec_signatures;
+				i->nsec_rrset.next = NULL;
+				return;
+			}
+		}
 		i->cur_node  = ldns_rbtree_next(i->cur_node);
 		i->cur_rrset = i->cur_node != LDNS_RBTREE_NULL
 		    ? ((ldns_dnssec_name *)i->cur_node->data)->rrsets
@@ -508,16 +763,18 @@ chase(ldns_dnssec_rrsets *rrset, ldns_dnssec_zone *support,
 			for (rrs = key_rrset->rrs; rrs; rrs = rrs->next)
 				if (ldns_rr_compare_ds(rr, rrs->rr))
 					break;
-			if (! rrs) {
-				s = LDNS_STATUS_CRYPTO_NO_DNSKEY;
-				break;
-			}
+			/* No DS found, try one of the other keys */
+			if (! rrs)
+				continue;
 		}
-		/* Pursue the chase with the verifying key (or its DS) */
+		/* Pursue the chase with the verifying key (or its DS)
+		 * and we're done.
+		 */
 		s = chase(key_rrset, support, support_keys, trusted);
-		if (s != 0)
-			break;
+		break;
 	}
+	if (i == ldns_rr_list_rr_count(verifying_keys))
+		s = LDNS_STATUS_CRYPTO_NO_DNSKEY;
 done_free_verifying_keys:
 	ldns_rr_list_free(verifying_keys);
 	return s;
@@ -528,9 +785,9 @@ done_free_verifying_keys:
  *
  */
 getdns_return_t
-getdns_validate_dnssec(struct getdns_list *records_to_validate,
-    struct getdns_list *support_records,
-    struct getdns_list *trust_anchors)
+getdns_validate_dnssec(getdns_list *records_to_validate,
+    getdns_list *support_records,
+    getdns_list *trust_anchors)
 {
 	getdns_return_t r;
 	ldns_rr_list     *trusted;
@@ -540,7 +797,7 @@ getdns_validate_dnssec(struct getdns_list *records_to_validate,
 	zone_iter i;
 	ldns_dnssec_rrsets *rrset;
 	ldns_dnssec_rrs *rrs;
-	ldns_status s = LDNS_STATUS_OK;
+	ldns_status s = LDNS_STATUS_ERR;
 
 	if ((r = priv_getdns_rr_list_from_list(trust_anchors, &trusted)))
 		return r;
@@ -572,8 +829,7 @@ getdns_validate_dnssec(struct getdns_list *records_to_validate,
 	for (rrset_iter_init_zone(&i, to_validate);
 	    (rrset = rrset_iter_value(&i)); rrset_iter_next(&i)) {
 
-		s |= chase(rrset, support, support_keys, trusted);
-		if (s != 0)
+		if ((s = chase(rrset, support, support_keys, trusted)))
 			break;
 	}
 	if (s == LDNS_STATUS_CRYPTO_BOGUS)
@@ -594,16 +850,25 @@ done_free_trusted:
 }				/* getdns_validate_dnssec */
 
 int
-priv_getdns_parse_ta_file(time_t *ta_mtime, ldns_rr_list *ta_rrs)
+priv_getdns_parse_ta_file(time_t *ta_mtime, getdns_list *ta_rrs)
 {
-	uint32_t ttl = 3600;
-	ldns_rdf* orig = NULL, *prev = NULL;
-	int line = 1;
-	ldns_status s;
-	ldns_rr *rr;
-	int nkeys;
+
+	struct gldns_file_parse_state pst;
 	struct stat st;
+	struct {
+		uint16_t id;
+		uint16_t flags;
+		uint16_t qdcount;
+		uint16_t ancount;
+		uint16_t nscount;
+		uint16_t arcount;
+		uint8_t rr[8192]; /* Reasonable max size for a single RR */
+	} pkt;
+	size_t len, dname_len;
 	FILE *in;
+	priv_getdns_rr_iter rr_iter;
+	getdns_dict *rr_dict = NULL;
+	int ta_count = 0;
 
 	if (stat(TRUST_ANCHOR_FILE, &st) != 0)
 		return 0;
@@ -611,58 +876,51 @@ priv_getdns_parse_ta_file(time_t *ta_mtime, ldns_rr_list *ta_rrs)
 	if (ta_mtime)
 		*ta_mtime = st.st_mtime;
 
-	in = fopen(TRUST_ANCHOR_FILE, "r");
-	if (!in)
+	if (!(in = fopen(TRUST_ANCHOR_FILE, "r")))
 		return 0;
 
-	nkeys = 0;
-	while (! feof(in)) {
-		rr = NULL;
-		s = ldns_rr_new_frm_fp_l(&rr, in, &ttl, &orig, &prev, &line);
-		if (s == LDNS_STATUS_SYNTAX_EMPTY /* empty line */
-		    || s == LDNS_STATUS_SYNTAX_TTL /* $TTL */
-		    || s == LDNS_STATUS_SYNTAX_ORIGIN /* $ORIGIN */)
-			continue;
+	pkt.id = pkt.flags = pkt.qdcount = pkt.nscount = pkt.arcount = 0;
+	pkt.ancount = htons(1);
 
-		if (s != LDNS_STATUS_OK) {
-			ldns_rr_free(rr);
-			nkeys = 0;
+	memset(&pst, 0, sizeof(pst));
+	pst.default_ttl = 3600;
+	pst.lineno = 1;
+
+	while (!feof(in)) {
+		len = sizeof(pkt.rr);
+		dname_len = 0;
+		if (gldns_fp2wire_rr_buf(in, pkt.rr, &len, &dname_len, &pst))
 			break;
-		}
-		if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_DS ||
-		    ldns_rr_get_type(rr) == LDNS_RR_TYPE_DNSKEY) {
-
-			nkeys++;
-			if (ta_rrs) {
-				ldns_rr_list_push_rr(ta_rrs, rr);
-				continue;
-			}
-		}
-		ldns_rr_free(rr);
+		if (len == 0)  /* empty, $TTL, $ORIGIN */
+			continue;
+		if (gldns_wirerr_get_type(pkt.rr, len, dname_len) 
+		    != LDNS_RR_TYPE_DS &&
+		    gldns_wirerr_get_type(pkt.rr, len, dname_len)
+		    != LDNS_RR_TYPE_DNSKEY)
+			continue;
+		if (!priv_getdns_rr_iter_init(&rr_iter, (void *)&pkt, sizeof(pkt)))
+			break;
+		if (!(rr_dict = priv_getdns_rr_iter2rr_dict(NULL, &rr_iter)))
+			break;
+		if (ta_rrs && getdns_list_append_dict(ta_rrs, rr_dict))
+			break;
+		getdns_dict_destroy(rr_dict);
+		rr_dict = NULL;
+		ta_count++;
 	}
-	ldns_rdf_deep_free(orig);
-	ldns_rdf_deep_free(prev);
+	if (rr_dict)
+		getdns_dict_destroy(rr_dict);
 	fclose(in);
-	return nkeys;
+
+	return ta_count;
 }
 
 getdns_list *
 getdns_root_trust_anchor(time_t *utc_date_of_anchor)
 {
-	getdns_list  *tas_gd_list = NULL;
-	ldns_rr_list *tas_rr_list = ldns_rr_list_new();
-
-	if (! tas_rr_list)
-		return NULL;
-
-	if (! priv_getdns_parse_ta_file(utc_date_of_anchor, tas_rr_list)) {
-		goto done_free_tas_rr_list;
-	}
-	tas_gd_list = create_list_from_rr_list(NULL, tas_rr_list);
-
-done_free_tas_rr_list:
-	ldns_rr_list_deep_free(tas_rr_list);
-	return tas_gd_list;
+	getdns_list *ta_rrs = getdns_list_create();
+	(void) priv_getdns_parse_ta_file(utc_date_of_anchor, ta_rrs);
+	return ta_rrs;
 }
 
 /* dnssec.c */
