@@ -81,20 +81,27 @@ handle_network_request_error(getdns_network_req * netreq, int err)
 void
 priv_getdns_check_dns_req_complete(getdns_dns_req *dns_req)
 {
-	getdns_network_req *netreq;
+	getdns_network_req **netreq_p, *netreq;
 	int results_found = 0;
 	
-	for (netreq = dns_req->first_req; netreq; netreq = netreq->next)
+	for (netreq_p = dns_req->netreqs; (netreq = *netreq_p); netreq_p++)
 		if (netreq->state != NET_REQ_FINISHED &&
 		    netreq->state != NET_REQ_CANCELED)
 			return;
-		else if (netreq->result)
+		else if (netreq->response_len > 0)
 			results_found = 1;
 
-	if (! results_found)
+	if (dns_req->internal_cb)
+		dns_req->internal_cb(dns_req);
+	else if (! results_found)
 		priv_getdns_call_user_callback(dns_req, NULL);
-	else if (is_extension_set(dns_req->extensions,
-	    "dnssec_return_validation_chain"))
+	else if (dns_req->dnssec_return_validation_chain
+#ifdef STUB_NATIVE_DNSSEC
+	    || (dns_req->context->resolution_type == GETDNS_RESOLUTION_STUB
+	        && (dns_req->dnssec_return_status ||
+	            dns_req->dnssec_return_only_secure))
+#endif
+	    )
 		priv_getdns_get_validation_chain(dns_req);
 	else
 		priv_getdns_call_user_callback(
@@ -131,14 +138,16 @@ submit_network_request(getdns_network_req *netreq)
 	getdns_return_t r;
 	getdns_dns_req *dns_req = netreq->owner;
 
-	if (dns_req->context->resolution_type == GETDNS_RESOLUTION_RECURSING ||
+	if (dns_req->context->resolution_type == GETDNS_RESOLUTION_RECURSING
 	    /* TODO: Until DNSSEC with the new async stub resolver is finished,
 	     *       use unbound when we need DNSSEC.
 	     */
-	    (dns_req->extensions && (
-	    is_extension_set(dns_req->extensions, "dnssec_return_status") ||
-	    is_extension_set(dns_req->extensions, "dnssec_return_only_secure") ||
-	    is_extension_set(dns_req->extensions, "dnssec_return_validation_chain")))) {
+#ifndef STUB_NATIVE_DNSSEC
+	    || dns_req->dnssec_return_status
+	    || dns_req->dnssec_return_only_secure
+	    || dns_req->dnssec_return_validation_chain
+#endif
+	    ) {
 
 		/* schedule the timeout */
 		if (! dns_req->timeout.timeout_cb) {
@@ -165,15 +174,15 @@ static getdns_return_t
 getdns_general_ns(getdns_context *context, getdns_eventloop *loop,
     const char *name, uint16_t request_type, getdns_dict *extensions,
     void *userarg, getdns_transaction_t *transaction_id,
-    getdns_callback_t callbackfn, int usenamespaces)
+    getdns_callback_t callbackfn, internal_cb_t internal_cb, int usenamespaces)
 {
 	getdns_return_t r = GETDNS_RETURN_GOOD;
-	getdns_network_req *netreq;
+	getdns_network_req *netreq, **netreq_p;
 	getdns_dns_req *req;
 	getdns_dict *localnames_response;
 	size_t i;
 
-	if (!context || !name || !callbackfn)
+	if (!context || !name || (!callbackfn && !internal_cb))
 		return GETDNS_RETURN_INVALID_PARAMETER;
 	
 	if ((r = validate_dname(name)))
@@ -193,6 +202,7 @@ getdns_general_ns(getdns_context *context, getdns_eventloop *loop,
 
 	req->user_pointer = userarg;
 	req->user_callback = callbackfn;
+	req->internal_cb = internal_cb;
 
 	if (transaction_id)
 		*transaction_id = req->trans_id;
@@ -201,7 +211,9 @@ getdns_general_ns(getdns_context *context, getdns_eventloop *loop,
 
 	if (!usenamespaces)
 		/* issue all network requests */
-		for (netreq = req->first_req; !r && netreq; netreq = netreq->next)
+		for ( netreq_p = req->netreqs
+		    ; !r && (netreq = *netreq_p)
+		    ; netreq_p++)
 			r = submit_network_request(netreq);
 
 	else for (i = 0; i < context->namespace_count; i++) {
@@ -221,11 +233,10 @@ getdns_general_ns(getdns_context *context, getdns_eventloop *loop,
 			   if this means we go onto the next namespace instead
 			   of returning */
 			r = GETDNS_RETURN_GOOD;
-			netreq = req->first_req;
-			while (!r && netreq) {
+			for ( netreq_p = req->netreqs
+			    ; !r && (netreq = *netreq_p)
+			    ; netreq_p++)
 				r = submit_network_request(netreq);
-				netreq = netreq->next;
-			}
 			break;
 		} else
 			r = GETDNS_RETURN_BAD_CONTEXT;
@@ -244,11 +255,11 @@ getdns_return_t
 priv_getdns_general_loop(getdns_context *context, getdns_eventloop *loop,
     const char *name, uint16_t request_type, getdns_dict *extensions,
     void *userarg, getdns_transaction_t *transaction_id,
-    getdns_callback_t callback)
+    getdns_callback_t callback, internal_cb_t internal_cb)
 {
 	return getdns_general_ns(context, loop,
 	    name, request_type, extensions,
-	    userarg, transaction_id, callback, 0);
+	    userarg, transaction_id, callback, internal_cb, 0);
 
 }				/* getdns_general_loop */
 
@@ -257,24 +268,28 @@ priv_getdns_address_loop(getdns_context *context, getdns_eventloop *loop,
     const char *name, getdns_dict *extensions, void *userarg,
     getdns_transaction_t *transaction_id, getdns_callback_t callback)
 {
-	int cleanup_extensions = 0;
+	getdns_dict *my_extensions = extensions;
 	getdns_return_t r;
+	uint32_t value;
 
-	if (!extensions) {
-		if (!(extensions = getdns_dict_create_with_context(context)))
+	if (!my_extensions) {
+		if (!(my_extensions=getdns_dict_create_with_context(context)))
 			return GETDNS_RETURN_MEMORY_ERROR;
-		cleanup_extensions = 1;
-	}
-	if ((r = getdns_dict_set_int(extensions, "return_both_v4_and_v6",
-	    GETDNS_EXTENSION_TRUE)))
+	} else if (
+	    getdns_dict_get_int(my_extensions, "return_both_v4_and_v6", &value)
+	    && (r = getdns_dict_copy(extensions, &my_extensions)))
 		return r;
-	
-	r = getdns_general_ns(context, loop,
-	    name, GETDNS_RRTYPE_A, extensions,
-	    userarg, transaction_id, callback, 1);
 
-	if (cleanup_extensions)
-		getdns_dict_destroy(extensions);
+	if (my_extensions != extensions && (r = getdns_dict_set_int(
+	    my_extensions, "return_both_v4_and_v6", GETDNS_EXTENSION_TRUE)))
+		return r;
+
+	r = getdns_general_ns(context, loop,
+	    name, GETDNS_RRTYPE_AAAA, my_extensions,
+	    userarg, transaction_id, callback, NULL, 1);
+
+	if (my_extensions != extensions)
+		getdns_dict_destroy(my_extensions);
 
 	return r;
 } /* getdns_address_loop */
@@ -287,7 +302,7 @@ priv_getdns_hostname_loop(getdns_context *context, getdns_eventloop *loop,
 	struct getdns_bindata *address_data;
 	struct getdns_bindata *address_type;
 	uint16_t req_type;
-	char *name;
+	char name[1024];
 	getdns_return_t retval;
 
 	if ((retval =
@@ -309,11 +324,60 @@ priv_getdns_hostname_loop(getdns_context *context, getdns_eventloop *loop,
 		req_type = GETDNS_RRTYPE_PTR;
 	else
 		return GETDNS_RETURN_INVALID_PARAMETER;
-	if ((name = reverse_address(address_data)) == NULL)
+
+	switch (address_data->size) {
+	case 4:
+		(void)snprintf(name, sizeof(name),
+		    "%hhu.%hhu.%hhu.%hhu.in-addr.arpa.",
+		    ((uint8_t *)address_data->data)[3],
+		    ((uint8_t *)address_data->data)[2],
+		    ((uint8_t *)address_data->data)[1],
+		    ((uint8_t *)address_data->data)[0]);
+		break;
+	case 16:
+		(void)snprintf(name, sizeof(name),
+		    "%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx."
+		    "%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx."
+		    "%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx."
+		    "%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.%hhx.ip6.arpa.",
+		    ((uint8_t *)address_data->data)[15] & 0x0F,
+		    ((uint8_t *)address_data->data)[15] >> 4,
+		    ((uint8_t *)address_data->data)[14] & 0x0F,
+		    ((uint8_t *)address_data->data)[14] >> 4,
+		    ((uint8_t *)address_data->data)[13] & 0x0F,
+		    ((uint8_t *)address_data->data)[13] >> 4,
+		    ((uint8_t *)address_data->data)[12] & 0x0F,
+		    ((uint8_t *)address_data->data)[12] >> 4,
+		    ((uint8_t *)address_data->data)[11] & 0x0F,
+		    ((uint8_t *)address_data->data)[11] >> 4,
+		    ((uint8_t *)address_data->data)[10] & 0x0F,
+		    ((uint8_t *)address_data->data)[10] >> 4,
+		    ((uint8_t *)address_data->data)[9] & 0x0F,
+		    ((uint8_t *)address_data->data)[9] >> 4,
+		    ((uint8_t *)address_data->data)[8] & 0x0F,
+		    ((uint8_t *)address_data->data)[8] >> 4,
+		    ((uint8_t *)address_data->data)[7] & 0x0F,
+		    ((uint8_t *)address_data->data)[7] >> 4,
+		    ((uint8_t *)address_data->data)[6] & 0x0F,
+		    ((uint8_t *)address_data->data)[6] >> 4,
+		    ((uint8_t *)address_data->data)[5] & 0x0F,
+		    ((uint8_t *)address_data->data)[5] >> 4,
+		    ((uint8_t *)address_data->data)[4] & 0x0F,
+		    ((uint8_t *)address_data->data)[4] >> 4,
+		    ((uint8_t *)address_data->data)[3] & 0x0F,
+		    ((uint8_t *)address_data->data)[3] >> 4,
+		    ((uint8_t *)address_data->data)[2] & 0x0F,
+		    ((uint8_t *)address_data->data)[2] >> 4,
+		    ((uint8_t *)address_data->data)[1] & 0x0F,
+		    ((uint8_t *)address_data->data)[1] >> 4,
+		    ((uint8_t *)address_data->data)[0] & 0x0F,
+		    ((uint8_t *)address_data->data)[0] >> 4);
+		break;
+	default:
 		return GETDNS_RETURN_INVALID_PARAMETER;
+	}
 	retval = priv_getdns_general_loop(context, loop, name, req_type,
-	    extensions, userarg, transaction_id, callback);
-	free(name);
+	    extensions, userarg, transaction_id, callback, NULL);
 	return retval;
 }				/* getdns_hostname_loop */
 
@@ -323,7 +387,7 @@ priv_getdns_service_loop(getdns_context *context, getdns_eventloop *loop,
     getdns_transaction_t * transaction_id, getdns_callback_t callback)
 {
 	return getdns_general_ns(context, loop, name, GETDNS_RRTYPE_SRV,
-	    extensions, userarg, transaction_id, callback, 1);
+	    extensions, userarg, transaction_id, callback, NULL, 1);
 }				/* getdns_service_loop */
 
 /**
@@ -338,7 +402,7 @@ getdns_general(getdns_context *context,
 	if (!context) return GETDNS_RETURN_INVALID_PARAMETER;
 	return priv_getdns_general_loop(context, context->extension,
 	    name, request_type, extensions,
-	    userarg, transaction_id, callback);
+	    userarg, transaction_id, callback, NULL);
 
 }				/* getdns_general */
 
