@@ -164,6 +164,18 @@ typedef struct getdns_tcp_state {
 
 } getdns_tcp_state;
 
+/* TODO[TLS]: change this name to getdns_transport when API updated*/
+typedef enum getdns_base_transport {
+	GETDNS_BASE_TRANSPORT_MIN   = 0,
+	GETDNS_BASE_TRANSPORT_NONE  = 0,
+	GETDNS_BASE_TRANSPORT_UDP,
+	GETDNS_BASE_TRANSPORT_TCP_SINGLE, /* To be removed? */
+	GETDNS_BASE_TRANSPORT_STARTTLS,   /* Define before TCP to allow fallback */
+	GETDNS_BASE_TRANSPORT_TCP,
+	GETDNS_BASE_TRANSPORT_TLS,
+	GETDNS_BASE_TRANSPORT_MAX
+} getdns_base_transport_t;
+
 /**
  * Request data
  **/
@@ -191,6 +203,8 @@ typedef struct getdns_network_req
 	/* For stub resolving */
 	struct getdns_upstream *upstream;
 	int                     fd;
+	getdns_base_transport_t dns_base_transports[GETDNS_BASE_TRANSPORT_MAX];
+	int                     transport;
 	getdns_eventloop_event  event;
 	getdns_tcp_state        tcp;
 	uint16_t                query_id;
